#include <stdio.h>
#include <stdlib.h>
#include <inttypes.h>
#include <stdint.h>
#include <string.h>

#define CHECK_ERROR(condition,message) \
do {if (condition) { \
    fprintf(stderr, "%s:%d:%s: %s\n",__FILE__,__LINE__,__func__,message); \
    abort(); \
}}while(0)

#define CHECK_ERRNO(condition, message) \
do {if (condition) { \
    fprintf(stderr, "%s:%d:%s:%s: ",__FILE__,__LINE__,__func__,message); \
    fflush(stderr); \
    perror(NULL); \
    abort(); \
}}while(0)

#define CHECK_FILE(condition,file,message) \
do {if (condition) { \
    fprintf(stderr, "%s:%d:%s:%s: ",__FILE__,__LINE__,__func__,message); \
    fflush(stderr); \
    if (feof(file)) { \
        fprintf(stderr,"unexpected EOF\n"); \
    } else { \
        perror(message); \
    } \
    abort(); \
}}while(0)

void dump(FILE *infile, FILE *outfile, long offset, size_t size);

void analyze(FILE *infile, long file_length);

int main(int argc, char **argv)
{
    printf("utf_tab 0.0\n\n");
    CHECK_ERROR(argc != 2, "Incorrect program usage\n\nusage: utf_tab file");

    /* open file */
    FILE *infile = fopen(argv[1], "rb");
    CHECK_ERRNO(!infile, "fopen");

    /* get file size */
    CHECK_ERRNO(fseek(infile, 0 , SEEK_END) != 0, "fseek");
    long file_length = ftell(infile);
    CHECK_ERRNO(file_length == -1, "ftell");

    rewind(infile);

    analyze(infile, file_length);

    exit(EXIT_SUCCESS);
}

void dump(FILE *infile, FILE *outfile, long offset, size_t size)
{
    unsigned char buf[0x800];

    CHECK_ERRNO(fseek(infile, offset, SEEK_SET) != 0, "fseek");

    while (size > 0)
    {
        size_t bytes_to_copy = sizeof(buf);
        if (bytes_to_copy > size) bytes_to_copy = size;

        size_t bytes_read = fread(buf, 1, bytes_to_copy, infile);
        CHECK_FILE(bytes_read != bytes_to_copy, infile, "fread");

        size_t bytes_written = fwrite(buf, 1, bytes_to_copy, outfile);
        CHECK_FILE(bytes_written != bytes_to_copy, outfile, "fwrite");

        size -= bytes_to_copy;
    }
}

uint32_t read_32_le(unsigned char bytes[4])
{
    uint32_t result = 0;
    for (int i=3; i>=0; i--) result = (result << 8) | bytes[i];
    return result;
}
uint16_t read_16_le(unsigned char bytes[2])
{
    uint32_t result = 0;
    for (int i=1; i>=0; i--) result = (result << 8) | bytes[i];
    return result;
}
uint32_t read_32_be(unsigned char bytes[4])
{
    uint32_t result = 0;
    for (int i=0; i<4; i++) result = (result << 8) | bytes[i];
    return result;
}
uint16_t read_16_be(unsigned char bytes[2])
{
    uint32_t result = 0;
    for (int i=0; i<2; i++) result = (result << 8) | bytes[i];
    return result;
}

uint8_t get_byte(FILE *infile)
{
    unsigned char buf[1];

    size_t bytes_read = fread(buf, 1, 1, infile);
    CHECK_FILE(bytes_read != 1, infile, "fread");

    return buf[0];
}
uint8_t get_byte_seek(long offset, FILE *infile)
{
    CHECK_ERRNO(fseek(infile, offset, SEEK_SET) != 0, "fseek");

    return get_byte(infile);
}

uint16_t get_16_be(FILE *infile)
{
    unsigned char buf[2];
    size_t bytes_read = fread(buf, 1, 2, infile);
    CHECK_FILE(bytes_read != 2, infile, "fread");

    return read_16_be(buf);
}
uint16_t get_16_be_seek(long offset, FILE *infile)
{
    CHECK_ERRNO(fseek(infile, offset, SEEK_SET) != 0, "fseek");

    return get_16_be(infile);
}
uint32_t get_32_be(FILE *infile)
{
    unsigned char buf[4];
    size_t bytes_read = fread(buf, 1, 4, infile);
    CHECK_FILE(bytes_read != 4, infile, "fread");

    return read_32_be(buf);
}
uint32_t get_32_be_seek(long offset, FILE *infile)
{
    CHECK_ERRNO(fseek(infile, offset, SEEK_SET) != 0, "fseek");

    return get_32_be(infile);
}

void analyze_utf(FILE *infile, long offset, int indent);

void analyze(FILE *infile, long file_length)
{
    long offset = 0;
    int indent = 0;

    analyze_utf(infile, offset, indent);
}

#define COLUMN_STORAGE_MASK         0xf0
#define COLUMN_STORAGE_PERROW       0x50
#define COLUMN_STORAGE_CONSTANT     0x30
#define COLUMN_STORAGE_ZERO         0x10

#define COLUMN_TYPE_MASK            0x0f
#define COLUMN_TYPE_DATA            0x0b
#define COLUMN_TYPE_STRING          0x0a
#define COLUMN_TYPE_FLOAT           0x08
#define COLUMN_TYPE_4BYTE           0x04
#define COLUMN_TYPE_2BYTE2          0x03
#define COLUMN_TYPE_2BYTE           0x02
#define COLUMN_TYPE_1BYTE2          0x01
#define COLUMN_TYPE_1BYTE           0x00

struct utf_column_info
{
    uint8_t type;
    const char *column_name;
    long constant_offset;
};

struct utf_table_info
{
    long table_offset;
    uint32_t table_size;
    uint32_t schema_offset;
    uint32_t rows_offset;
    uint32_t string_table_offset;
    uint32_t data_offset;
    const char *string_table;
    const char *table_name;
    uint16_t columns;
    uint16_t row_width;
    uint32_t rows;

    const struct utf_column_info *schema;
};

void fprintf_table_info(FILE *outfile, const struct utf_table_info *table_info, int indent);

#define INDENT_LEVEL 2
void fprintf_indent(FILE *outfile, int indent)
{
    fprintf(outfile, "%*s",indent,"");
}

void analyze_utf(FILE *infile, const long offset, int indent)
{
    unsigned char buf[4];
    struct utf_table_info table_info;
    char *string_table = NULL;
    struct utf_column_info * schema = NULL;

    fprintf_indent(stdout, indent);
    printf("{\n");

    indent += INDENT_LEVEL;

    table_info.table_offset = offset;

    /* seek to header */
    CHECK_ERRNO(fseek(infile, offset, SEEK_SET) != 0, "fseek");

    /* read header */
    size_t bytes_read = fread(buf, 1, 4, infile);
    CHECK_FILE(bytes_read != 4, infile, "fread");

    /* check header */
    static const char UTF_signature[4] = "@UTF"; /* intentionally unterminated */
    if (memcmp(buf, UTF_signature, sizeof(UTF_signature)))
    {
        fprintf_indent(stdout, indent);
        printf("not a @UTF table at %08" PRIx32 "\n", (uint32_t)offset);
        goto cleanup;
    }

    /* get table size */
    table_info.table_size = get_32_be(infile);

    table_info.schema_offset = 0x20;
    table_info.rows_offset = get_32_be(infile);
    table_info.string_table_offset = get_32_be(infile);
    table_info.data_offset = get_32_be(infile);
    const uint32_t table_name_string = get_32_be(infile);
    table_info.columns = get_16_be(infile);
    table_info.row_width = get_16_be(infile);
    table_info.rows = get_32_be(infile);

    /* allocate for string table */
    const int string_table_size =
        table_info.data_offset-table_info.string_table_offset;
    string_table = malloc(string_table_size+1);
    CHECK_ERRNO(!string_table, "malloc");
    table_info.string_table = string_table;
    memset(string_table, 0, string_table_size+1);

    /* load schema */
    schema = malloc(sizeof(struct utf_column_info) * table_info.columns);
    CHECK_ERRNO(!schema, "malloc");
    {
        int i;
        for (i = 0; i < table_info.columns; i++)
        {
            schema[i].type = get_byte(infile);
            schema[i].column_name = string_table + get_32_be(infile);

            if ((schema[i].type & COLUMN_STORAGE_MASK) == COLUMN_STORAGE_CONSTANT)
            {
                schema[i].constant_offset = ftell(infile);
                switch (schema[i].type & COLUMN_TYPE_MASK)
                {
                    case COLUMN_TYPE_STRING:
                        get_32_be(infile);
                        break;
                    case COLUMN_TYPE_DATA:
                        get_32_be(infile);
                        get_32_be(infile);
                        break;
                    case COLUMN_TYPE_FLOAT:
                    case COLUMN_TYPE_4BYTE:
                        get_32_be(infile);
                        break;
                    case COLUMN_TYPE_2BYTE2:
                    case COLUMN_TYPE_2BYTE:
                        get_16_be(infile);
                        break;
                    case COLUMN_TYPE_1BYTE2:
                    case COLUMN_TYPE_1BYTE:
                        get_byte(infile);
                        break;
                    default:
                        CHECK_ERROR(1, "unknown type for constant");
                }
            }
        }
    }

    table_info.schema = schema;

    /* seek to string table */
    CHECK_ERRNO(fseek(infile, table_info.string_table_offset+8+offset, SEEK_SET) != 0, "fseek");

    /* read string table */
    bytes_read = fread(string_table, 1, string_table_size, infile);
    CHECK_FILE(bytes_read != string_table_size, infile, "fread");
    table_info.table_name = table_info.string_table+table_name_string;

#if 0
    fprintf_table_info(stdout, &table_info, indent);
#endif

    /* explore the values */
    {
        int i, j;

        for (i = 0; i < table_info.rows; i++)
        {
            uint32_t row_offset =
                table_info.table_offset + 8 + table_info.rows_offset +
                i * table_info.row_width;
            const uint32_t row_start_offset = row_offset;

            fprintf_indent(stdout, indent);
            printf("%s[%d] = {\n", table_info.table_name, i);
            indent += INDENT_LEVEL;
            for (j = 0; j < table_info.columns; j++)
            {
                uint8_t type = table_info.schema[j].type;
                long constant_offset = table_info.schema[j].constant_offset;
                int constant = 0;
                fprintf_indent(stdout, indent);
#if 0
                printf("%08x %02x %s = ", row_offset-row_start_offset, type, table_info.schema[j].column_name);
#else
                printf("%s = ", table_info.schema[j].column_name);
#endif

                switch (schema[j].type & COLUMN_STORAGE_MASK)
                {
                    case COLUMN_STORAGE_PERROW:
                        break;
                    case COLUMN_STORAGE_CONSTANT:
                        constant = 1;
                        break;
                    case COLUMN_STORAGE_ZERO:
                        printf("constant no-storage 0\n");
                        continue;
                    default:
                        CHECK_ERROR(1, "unknown storage class");
                }

                if (1)
                {
                    long data_offset;
                    int bytes_read;

                    if (constant)
                    {
                        data_offset = constant_offset;
                        printf("constant ");
                    }
                    else
                    {
                        data_offset = row_offset;
                    }

                    switch (type & COLUMN_TYPE_MASK)
                    {
                        case COLUMN_TYPE_STRING:
                            {
                                uint32_t string_offset;
                                string_offset = get_32_be_seek(data_offset, infile);
                                bytes_read = 4;
                                printf("%s\n", table_info.string_table + string_offset);
                            }
                            break;
                        case COLUMN_TYPE_DATA:
                            {
                                uint32_t vardata_offset, vardata_size;

                                vardata_offset = get_32_be_seek(data_offset, infile);
                                vardata_size = get_32_be(infile);
                                bytes_read = 8;
                                printf("[0x%08" PRIx32 "]", vardata_offset);
                                printf(" (size 0x%08" PRIx32 ")\n", vardata_size);

                                if (vardata_size != 0)
                                {
                                    /* assume that the data is another table */
                                    analyze_utf(infile,
                                            table_info.table_offset + 8 +
                                            table_info.data_offset +
                                            vardata_offset,
                                            indent
                                            );
                                }
                            }
                            break;

                        case COLUMN_TYPE_4BYTE:
                            printf("%" PRId32 "\n",
                                    get_32_be_seek(data_offset, infile));
                            bytes_read = 4;
                            break;
                        case COLUMN_TYPE_2BYTE2:
                            printf("type 2 ");
                        case COLUMN_TYPE_2BYTE:
                            printf("%" PRId16 "\n",
                                    get_16_be_seek(data_offset, infile));
                            bytes_read = 2;
                            break;
                        case COLUMN_TYPE_FLOAT:
                            if (sizeof(float) == 4)
                            {
                                union {
                                    float float_value;
                                    uint32_t int_value;
                                } int_float;

                                int_float.int_value = get_32_be_seek(data_offset, infile);
                                printf("%f\n", int_float.float_value);
                            }
                            else
                            {
                                get_32_be_seek(data_offset, infile);
                                printf("float\n");
                            }
                            bytes_read = 4;
                            break;
                        case COLUMN_TYPE_1BYTE2:
                            printf("type 2 ");
                        case COLUMN_TYPE_1BYTE:
                            printf("%" PRId8 "\n",
                                    get_byte_seek(data_offset, infile));
                            bytes_read = 1;
                            break;
                        default:
                            CHECK_ERROR(1, "unknown normal type");
                    }

                    if (!constant)
                    {
                        row_offset += bytes_read;
                    }
                }
            }
            indent -= INDENT_LEVEL;
            fprintf_indent(stdout,indent);
            printf("}\n");

            CHECK_ERROR(row_offset - row_start_offset != table_info.row_width,
                    "column widths do now add up to row width");
        } /* row for loop end */
    } /* explore values block end */

cleanup:
    indent -= INDENT_LEVEL;
    fprintf_indent(stdout, indent);
    printf("}\n");

    if (string_table)
    {
        free(string_table);
        string_table = NULL;
    }

    if (schema)
    {
        free(schema);
        schema = NULL;
    }
}

void fprintf_table_info(FILE *outfile, const struct utf_table_info *table_info, int indent)
{
    fprintf_indent(outfile,indent);
    fprintf(outfile, "table: %s\n", table_info->table_name);
    fprintf_indent(outfile,indent);
    fprintf(outfile, "at offset           0x%08" PRIx32 " (size 0x%08" PRIx32 ")\n",
            (uint32_t)table_info->table_offset,
            table_info->table_size);
    fprintf_indent(outfile,indent);
    fprintf(outfile, "schema offset       0x%08" PRIx32 " (size 0x%08" PRIx32 ")\n",
            table_info->schema_offset,
            table_info->rows_offset-table_info->schema_offset);
    fprintf_indent(outfile,indent);
    fprintf(outfile, "rows offset         0x%08" PRIx32 " (size 0x%08" PRIx32 ")\n",
            table_info->rows_offset,
            table_info->string_table_offset-table_info->rows_offset);
    fprintf_indent(outfile,indent);
    fprintf(outfile, "string table offset 0x%08" PRIx32 " (size 0x%08" PRIx32 ")\n",
            table_info->string_table_offset,
            table_info->data_offset-table_info->string_table_offset);
    fprintf_indent(outfile,indent);
    fprintf(outfile, "data offset         0x%08" PRIx32 " (0x%08" PRIx32 ")\n",
            table_info->data_offset,
            table_info->table_size-table_info->data_offset);
    fprintf_indent(outfile,indent);
    fprintf(outfile, "columns:            %" PRIu32 "\n",
            table_info->columns);
    fprintf_indent(outfile,indent);
    fprintf(outfile, "rows:               %" PRIu32 "\n",
            table_info->rows);
    fprintf_indent(outfile,indent);
    fprintf(outfile, "row width:          %#" PRIx32 " bytes\n",
            table_info->row_width * table_info->rows);
}
